import argparse

from src.ucyph import __version__ as version
from src.ucyph.ciphers import *
from src.ucyph.utils import read_file, write_file

EXAMPLES = """
EXAMPLES
Encrypt input.txt using the Caesar cipher and save the result to output.txt:
```ucyph 3 input.txt -o output.txt``` 

Decrypt input.txt using the Vigenere cipher and save the result to output.txt: 
```ucyph -d 5 input.txt -k mypassword -o output.txt```

Encrypt input.txt using the Playfair cipher and overwrite the original file:
```ucyph 11 input.txt -k mypassword```
"""
USAGE_CODES = """
USAGE CODES
| Cipher   | Usage Code  | Requires Key |
|----------|-------------|--------------|
| Caesar   | 3           | No           |
| Vigenere | 5           | Yes          |
| Playfair | 11          | Yes          |
| Rot-13   | 13          | No           |
| Rot-47   | 47          | No           |
"""
CIPHER_DESCRIPTIONS = """
CIPHER DESCRIPTIONS
Caesar:   substitution cipher that shifts letters by 3 places
Rot-13:   substitution cipher that shifts letters by 13 places 
Rot-47:   substitution cipher that shifts letters, numbers, and symbols by 47 places
Vigenere: series of Caesar ciphers based on a keyword
Playfair: substitution cipher that pairs letters in a 5x5 grid and uses a keyword to encode/decode
"""
FUNCTION_MAP = {'3': caesar,
                '47': rot47,
                '13': rot13,
                '5': vigenere,
                '11': playfair, }


def func_names(a):
    """
    Return the name of the cipher function
    :param a: The usage code of the cipher
    :return: The name of the cipher function
    """
    FUNCTION_NAMES = {'3': 'Caesar',
                      '13': 'Rot-13',
                      '47': 'Rot-47',
                      '5': 'Vigenere',
                      '11': 'Playfair'}
    return FUNCTION_NAMES[a]


def interactive_mode():
    """
    Run the program in interactive mode
    :return: None
    """
    print(f"Interactive Mode: Follow the prompts to encrypt or decrypt text. \nUse Ctrl+C to exit.\n{USAGE_CODES}")
    try:
        while True:
            cipher = input("Select a cipher by entering the corresponding usage code. \nUsage Code: ")
            if cipher not in FUNCTION_MAP.keys():
                print("Invalid usage code.")
                continue
            text = input("Enter the text: ")
            key = input("Enter the key: ") if cipher in ['5', '11'] else None
            encode = input("Encrypt or decrypt? (e/d): ") != 'd' if cipher in ['3', '5', '11'] else True
            result = FUNCTION_MAP[cipher](text, key, encode)
            print(f"The {('encrypted' if encode else 'decrypted')} text is: {result}")
            if input("Continue? (y/n): ") != 'y':
                break
    except KeyboardInterrupt:
        print("\nExiting interactive mode.")


def parse_args():
    """
    Parse the command line arguments
    :return: The parsed arguments
    """

    parser = argparse.ArgumentParser(description='Encrypt or decrypt a file using historical ciphers',
                                     epilog=f'{EXAMPLES}\n{USAGE_CODES}\n{CIPHER_DESCRIPTIONS}',
                                     formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('code', nargs='?', choices=FUNCTION_MAP.keys(), metavar='<usage_code>',
                        help='usage code for the cipher you want to use. Refer to the USAGE CODES table below')
    parser.add_argument('file', nargs='?', metavar='<input_file>', help='input file to be encrypted/decrypted')
    parser.add_argument('-o', '--output', metavar='<output_file>',
                        help='file to output text to, defaults to input file if not specified')
    parser.add_argument('-k', '--key', metavar='<cipher_key>',
                        help='key/password for the cipher. Not required for some ciphers. Refer to the USAGE CODES table below')

    en_de = parser.add_mutually_exclusive_group()
    en_de.add_argument('-d', '--decode', action='store_true',
                       help='decode the string using current cipher. Cannot be used with -e/--encode')
    en_de.add_argument('-e', '--encode', action='store_true',
                       help='encode the string using current cipher. Cannot be used with -d/--decode. This is the default behavior')

    parser.add_argument('-i', '--interactive', action='store_true',
                        help='use interactive mode. cannot be used with other arguments')
    parser.add_argument('-v', '--version', action='version', version=f'%(prog)s {version}')

    return parser.parse_args()


def main():
    try:
        args = parse_args()
        if args.interactive or args.code is None or args.file is None:
            interactive_mode()
            return
        func = FUNCTION_MAP[args.code]
        encode = args.encode if args.encode or args.decode else True

        if args.code in ['5', '11'] and args.key is None:
            raise argparse.ArgumentError(None,
                                         f"The -k/--key argument is required for the selected cipher: {func_names(args.code)}"
                                         f"\n\nUsage: ucyph {args.code} {args.file} -k <key> [-o <output_file>] [-d/-e]")

        text = read_file(args.file)
        final = func(text, args.key, encode) if args.key else func(text, encode)

        out_file = args.output if args.output else args.file
        write_file(final, out_file)

        crypt = 'decrypted' if args.decode else 'encrypted'
        print(f'The {crypt} text has been saved to \"{out_file}\"')

    except FileNotFoundError as e:
        print(f'Error reading from input file: {e}')
    except PermissionError as e:
        print(f'Permission denied: {e}')
    except argparse.ArgumentError as e:
        print(f'Error: {e}')
    except Exception as e:
        print(f'An unexpected error occurred: {e}')


if __name__ == '__main__':
    main()



# ---- Ciphers ----

# OG Caesar - shift of 3
def caesar(text: str, encode=True, *args) -> str:
    """
    The Caesar cipher is one of the earliest known and simplest ciphers. It is a type of substitution cipher in which
    each letter in the plaintext is 'shifted' a certain number of places down the alphabet. For example, with a shift
    of 3, A would be replaced by D, B would become E, and so on.
    :param text: The string to be encoded/decoded
    :param encode: Whether to encode or decode the string
    :return: The encoded/decoded string
    """
    text = text.lower()
    az = 'abcdefghijklmnopqrstuvwxyz'
    new = ''
    if encode:
        for i in text:
            if i.isalpha():
                new += az[(az.index(i) + 3) % 26]
            else:
                new += i
    else:
        for i in text:
            if i.isalpha():
                if az.index(i) < 3:
                    new += az[26 - (abs(az.index(i) - 3) % 26)]
                else:
                    new += az[az.index(i) - 3]
            else:
                new += i
    return new


# ROT-47
def rot47(text: str, *args) -> str:
    """
    ROT-47 is a derivative of ROT-13 which, in addition to scrambling the basic letters, also treats numbers and
    common symbols. ROT-47 is frequently used to obfuscate plaintext in online forums such as Usenet to hide
    spoilers, punchlines, and puzzle solutions from the casual glance.
    :param text: The string to be encoded/decoded
    :return: The encoded/decoded string
    """
    key = '!"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\\]^_`abcdefghijklmnopqrstuvwxyz{|}~'
    words = text.split(' ')
    final = []

    for word in words:
        crypted = []
        for letter in word:
            crypted.append(key[(key.index(letter) + 47) % 94])
        final.append(''.join(crypted))

    return ' '.join(final)


# ROT-13
def rot13(text: str, *args) -> str:
    """
    ROT-13 ("rotate by 13 places", sometimes hyphenated ROT-13) is a simple letter substitution cipher that replaces a
    letter with the 13th letter after it, in the alphabet. ROT13 is a special case of the Caesar cipher which was
    developed in ancient Rome.
    :param text: The string to be encoded/decoded
    :return: The encoded/decoded string
    """
    az = 'abcdefghijklmnopqrstuvwxyz'
    decrypted = ''
    for i in text:
        if i.isalpha():
            if i.islower():
                if az.index(i) < 13:
                    decrypted += az[az.index(i) + 13]
                elif az.index(i) >= 13:
                    decrypted += (az[(az.index(i) + 13) % 13])
            else:
                if az.upper().index(i) < 13:
                    decrypted += az.upper()[az.upper().index(i) + 13]
                elif az.upper().index(i) >= 13:
                    decrypted += az.upper()[(az.upper().index(i) + 13) % 13]
        else:
            decrypted += i
    return decrypted


# Vigenere
def vigenere(text: str, key: str, encode=True) -> str:
    """
    The Vigenère cipher is a method of encrypting alphabetic text by using a series of interwoven Caesar ciphers,
    based on the letters of a keyword. It was invented in 1553 by Giovan Battista Bellaso, and is thus
    also known as the Bellaso cipher. While the cipher is easy to understand and implement, for three centuries
    it resisted all attempts to break it; this earned it the description le chiffre indéchiffrable (French for
    'the indecipherable cipher').
    :param text: The string to be encoded/decoded
    :param key: The key to be used for the cipher
    :param encode: Whether to encode or decode the string
    :return: The encoded/decoded string
    """
    text = text.lower()
    key = key.lower()
    alph = 'abcdefghijklmnopqrstuvwxyz'
    key_repeat = []
    encoded = ''

    if encode:
        for position, value in enumerate(text):
            key_repeat.append(key[position % len(key)])

            if value in alph:
                alph_position = alph.index(value)
                key_position = alph.index(key_repeat[position])
                encoded += alph[(alph_position + key_position) % len(alph)]
            else:
                encoded += value

        return encoded

    else:
        for position, value in enumerate(text):
            key_repeat.append(key[position % len(key)])

            if value in alph:
                alph_position = alph.index(value)
                key_position = alph.index(key_repeat[position])
                if alph_position >= key_position:
                    encoded += alph[abs(alph_position - key_position) % len(alph)]
                else:
                    encoded += alph[len(alph) - abs(alph_position - key_position)]
            else:
                encoded += value
        return encoded


# Playfair
def playfair(text: str, key: str, encode=True) -> str:
    """
    The Playfair cipher is a digraph substitution cipher that was invented in 1854 by Charles Wheatstone,
    but named after Lord Playfair who heavily promoted its use.
    This cipher was used by the British during the Boer War and World War I and the Australians during World War II.

    :param text: The string to be encoded/decoded
    :param key: The key to be used for the cipher
    :param encode: Whether to encode or decode the string
    :return: The encoded/decoded string
    """
    AZ = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    diagram = []
    row = []

    # -----Diagram-----

    # Clean text and Convert key to upper + no space
    text = ''.join(char if char.isalpha() or char == ' ' else '' for char in text)
    key = key.upper()
    key = key.replace(' ', '')

    # Create beginning of diagram using key
    for letter in key:
        if letter == 'J': letter = 'I'

        if len(row) < 5 and letter in AZ:
            row.append(letter)
            AZ = AZ.replace(letter, '')
        elif letter in AZ:
            diagram.append(row)
            row = []
            row.append(letter)
            AZ = AZ.replace(letter, '')

    # Fill in rest of diagram using remaining AZ
    for letter in AZ:
        if letter == 'J': letter = 'I'

        if len(row) < 5 and letter in AZ:
            row.append(letter)
            AZ = AZ.replace(letter, '')
        elif letter in AZ:
            diagram.append(row)
            row = []
            row.append(letter)
            AZ = AZ.replace(letter, '')
    # Append last row to diagram
    diagram.append(row)

    # -----Format Text-----

    # Convert text to upper + no space
    text = text.upper()
    text = text.replace(' ', '')

    # Split text into pairs
    # If pair is two of the same letter,
    # Add 'X' between them
    pair = []
    plain_text = []
    for letter in text:
        if len(pair) == 1:
            if pair[0] == letter:
                pair.append('X')
                plain_text.append(pair)
                pair = [letter]
            else:
                pair.append(letter)
                plain_text.append(pair)
                pair = []
        else:
            pair.append(letter)
    # Check if there is a leftover letter in pair.
    # If so, add 'X' and append to plain_text
    if len(pair) == 1:
        pair.append('X')
        plain_text.append(pair)

    # -----Encode-----
    encoded = []
    for pair in plain_text:
        coord_a = []
        coord_b = []
        for row in diagram:
            if pair[0] in row:
                coord_a.append(diagram.index(row))
                coord_a.append(diagram[diagram.index(row)].index(pair[0]))
            if pair[1] in row:
                coord_b.append(diagram.index(row))
                coord_b.append(diagram[diagram.index(row)].index(pair[1]))
        # Rule 1: If they are on same row, add/sub 1 from row
        if coord_a[0] == coord_b[0]:
            if encode:
                encoded.append(diagram[coord_a[0]][(coord_a[1] + 1) % 5])
                encoded.append(diagram[coord_b[0]][(coord_b[1] + 1) % 5])
            else:
                # is the col index 0?
                if coord_a[1] == 0:
                    encoded.append(diagram[coord_a[0]][4])
                else:
                    encoded.append(diagram[coord_a[0]][coord_a[1] - 1])
                if coord_b[1] == 0:
                    encoded.append(diagram[coord_b[0]][4])
                else:
                    encoded.append(diagram[coord_b[0]][coord_b[1] - 1])
        # Rule 2: If they are on the same col, add/sub 1 from col
        elif coord_a[1] == coord_b[1]:
            if encode:
                encoded.append(diagram[(coord_a[0] + 1) % 5][coord_a[1]])
                encoded.append(diagram[(coord_b[0] + 1) % 5][coord_b[1]])
            else:
                # is the row index 0?
                if coord_a[0] == 0:
                    encoded.append(diagram[4][coord_a[1]])
                else:
                    encoded.append(diagram[coord_a[0] - 1][coord_a[1]])
                if coord_b[0] == 0:
                    encoded.append(diagram[4][coord_b[1]])
                else:
                    encoded.append(diagram[coord_b[0] - 1][coord_b[1]])
        # Rule 3: Make square and grab opposite corner on same row
        else:
            encoded.append(diagram[coord_a[0]][coord_b[1]])
            encoded.append(diagram[coord_b[0]][coord_a[1]])

    return ''.join(encoded)



from unittest.mock import patch, call
from src.ucyph.ucyph import interactive_mode


@patch('builtins.input')
@patch('builtins.print')
def test_interactive_mode(mock_print, mock_input):
    mock_input.side_effect = ['3', 'Hello', 'e', 'n']
    interactive_mode()
    mock_print.assert_has_calls([call("The encrypted text is: ebiil")])


@patch('builtins.input')
@patch('builtins.print')
def test_interactive_mode_with_key(mock_print, mock_input):
    mock_input.side_effect = ['5', 'Hello', 'mypassword', 'e', 'n']
    interactive_mode()
    mock_print.assert_has_calls([call("The encrypted text is: tcalg")])


@patch('builtins.input')
@patch('builtins.print')
def test_interactive_mode_invalid_usage_code(mock_print, mock_input):
    mock_input.side_effect = ['100', KeyboardInterrupt]
    interactive_mode()
    mock_print.assert_has_calls([call("Invalid usage code.")])



import os
import subprocess


def test_encode_caesar_cipher():
    result = subprocess.run(['python3', '-m', 'src.ucyph.ucyph', '3', 'tests/test_files/test_input.txt', '-e', '-k', 'key', '-o', 'tests/test_files/test_output.txt'],
                            capture_output=True, text=True)

    assert result.stdout == 'The encrypted text has been saved to "tests/test_files/test_output.txt"\n'
    with open('tests/test_files/test_output.txt', 'r') as f:
        file_contents = f.read()
    assert file_contents == 'khoor'
    os.remove('tests/test_files/test_output.txt')
